from __future__ import annotations

import abc
import logging
from pathlib import Path
from typing import Literal, TypeVar, Generic, TypedDict, Any, overload, Iterable

import numpy as np
import polars as pl
from polars.polars import ColumnNotFoundError
from typing_extensions import Self

from ._type import PathLike
from ._util import printdf
from .event import RigEvent, CamEvent
from .preference import PreferenceDict, load_preferences
from .session import Session, SessionInfo

__all__ = [
    'STIMPY_SOURCE_VERSION',
    'LOG_SUFFIX',
    'CAMERA_TYPE',
    #
    'RigConfig',
    'AbstractLog',
    'AbstractStimlog',
    #
    'AbstractStimProtocol',
    #
    'AbstractStimulusPattern'
]

STIMPY_SOURCE_VERSION = Literal['pyvstim', 'stimpy-bit', 'stimpy-git', 'debug']
LOG_SUFFIX = Literal['.log', '.riglog']
CAMERA_TYPE = Literal['facecam', 'eyecam', '1P_cam']

#
R = TypeVar('R', bound='AbstractLog')  # Riglog-Like
S = TypeVar('S', bound='AbstractStimlog')  # Stimlog-Like
P = TypeVar('P', bound='AbstractStimProtocol')  # protocol file

#
logger = logging.getLogger(__name__)


# =========== #
# Baselog ABC #
# =========== #

class RigConfig(TypedDict, total=False):
    source_version: STIMPY_SOURCE_VERSION
    """stimpy source version {'pyvstim', 'stimpy-bit', 'stimpy-git'}"""
    version: float
    """acquisition flag. i.e., 0.3"""
    commit_hash: str
    """git commit hash"""
    codes: dict[str, int]
    """<EVENT_TYPES>:<NUMBER>"""
    fields: tuple[str, ...]
    """column repr for the logging"""


class AbstractLog(Generic[S, P], metaclass=abc.ABCMeta):
    """ABC class for different stimpy/pyvstim log files. i.e., .log, .riglog"""

    log_config: RigConfig
    """config dict for the log file"""

    def __init__(self, root_path: PathLike,
                 log_suffix: LOG_SUFFIX,
                 diode_offset: bool = True,
                 reset_mapping: dict[int, list[str]] | None = None):
        """
        :param root_path: log file path or log directory
        :param log_suffix: log file suffix
        :param diode_offset: whether do the diode offset
        :param reset_mapping: Customized mapping
        """

        if not isinstance(root_path, Path):
            root_path = Path(root_path)

        if root_path.is_dir():
            self.riglog_file = self._find_logfile(root_path, log_suffix)
        else:
            self.riglog_file = root_path

        # set when get config
        self._with_square_brackets: bool = True
        self._skip_rows: int = 0

        self.log_config = self._get_log_config()
        self.version = self.log_config['source_version']
        self.dat = self._cache_asarray(self.riglog_file, self._with_square_brackets)

        #
        self._diode_offset = diode_offset
        self._reset_mapping = reset_mapping

    @classmethod
    def _find_logfile(cls, root: Path,
                      log_suffix: LOG_SUFFIX) -> Path:

        f = list(root.glob(f'*{log_suffix}'))
        if len(f) == 1:
            return f[0]

        elif len(f) == 0:
            logger.warning(f'no riglog under {root}, try to find in the subfolder...')
            for s in root.iterdir():
                if s.is_dir() and s.name.startswith('run0'):
                    try:
                        return cls._find_logfile(s, log_suffix)
                    except FileNotFoundError:
                        pass

            raise FileNotFoundError(f'no riglog file {log_suffix} under {root}')

        else:
            raise FileNotFoundError(f'more than one riglog files under {root}')

    # noinspection PyTypedDict
    def _get_log_config(self) -> RigConfig:
        """get config dict and source version for different logs from the # headers"""
        ret = RigConfig()
        with open(self.riglog_file) as f:
            for line in f:
                if '#' in line:
                    if 'RIG VERSION' in line:
                        ret['version'] = float(line.split(': ')[-1])
                        logger.debug(f'Parsed version: {ret["version"]}')

                    elif 'RIG GIT COMMIT HASH' in line:
                        ret['commit_hash'] = line.split(': ')[-1].strip()
                        logger.debug(f'Parsed commit_hash: {ret["commit_hash"]}')

                    elif 'CODES' in line:
                        codes = {}
                        content = line.replace('# CODES: ', '').strip()
                        iter_codes = content.split(',')
                        for pair in iter_codes:
                            code, num = pair.split('=')
                            code = code.strip()
                            value = int(num.strip())
                            codes[code.lower()] = value

                        ret['codes'] = codes
                        logger.debug(f'Parsed codes: {ret["codes"]}')

                    elif 'RIG CSV' in line:
                        content = line.replace('# RIG CSV: ', '').strip()
                        ret['fields'] = tuple(content.split(','))
                        logger.debug(f'Parsed fields: {ret["fields"]}')

                # 2025 new update log...
                elif '#' not in line and 'code' in line:
                    ret['fields'] = tuple(line.split(','))
                    self._skip_rows = 1
                    self._with_square_brackets = False
                    logger.debug(f'github stimpy version parsed: fields: {ret["fields"]}, without square brackets')

        # infer
        if 'opto' not in ret['codes']:
            ret['source_version'] = 'stimpy-bit'
            logger.info(f'Source version inferred by *opto* field: {ret["source_version"]}')

        if 'version' not in ret:
            ret['source_version'] = 'stimpy-git'
            logger.info(f'Source version inferred by config: {ret["source_version"]}')

        if 'opto' in ret['codes'] and ret['codes']['opto'] == 15:
            ret['source_version'] = 'pyvstim'
            logger.info(f'Source version inferred by *opto* field: {ret["source_version"]}')

        return ret

    @classmethod
    @abc.abstractmethod
    def _cache_asarray(cls, filepath: Path, square_brackets: bool) -> np.ndarray:
        pass

    # ============ #
    # RIGLOG EVENT #
    # ============ #

    def _event(self, code: int) -> np.ndarray:
        """shape(sample, 2) with time and value"""
        x = self.dat[:, 0] == code
        t = self.dat[x, 2].copy() / 1000  # s
        v = self.dat[x, 3].copy()
        return np.vstack([t, v]).T

    @property
    def exp_start_time(self) -> float:
        """experimental start time (in sec)"""
        return float(self.dat[0, 2].copy() / 1000)

    @property
    def exp_end_time(self) -> float:
        """experimental end time (in sec)"""
        return float(self.dat[-1, 2].copy() / 1000)

    @property
    def total_duration(self) -> float:
        """experimental duration (in sec)"""
        return self.exp_end_time - self.exp_start_time

    @property
    def screen_event(self) -> RigEvent:
        """screen rig event. i.e., diode pulse"""
        return RigEvent('screen', self._event(0))

    @property
    def imaging_event(self) -> RigEvent:
        """imaging rig event. i.e., 2photon pulse"""
        return RigEvent('imaging', self._event(1))

    @property
    def position_event(self) -> RigEvent:
        """position rig event. i.e., encoder pulse"""
        return RigEvent('position', self._event(2))

    @property
    def lick_event(self) -> RigEvent:
        """lick rig event. i.e., lick meter pulse"""
        return RigEvent('lick', self._event(3))

    @property
    def reward_event(self) -> RigEvent:
        """reward rig event. i.e., reward given pulse from lick meter"""
        if self.version == 'stimpy-git':
            return RigEvent('reward', self._event(5))
        else:
            return RigEvent('reward', self._event(4))

    @property
    def lap_event(self) -> RigEvent:
        """lap rig event. i.e., optic sensing for the reflective taps"""
        if self.version == 'stimpy-git':
            return RigEvent('lap', self._event(6))
        else:
            return RigEvent('lap', self._event(5))

    @property
    def act_event(self) -> RigEvent:
        """todo"""
        raise NotImplementedError('')

    @property
    def opto_event(self) -> RigEvent:
        """todo"""
        raise NotImplementedError('')

    class CameraEvent:
        """camera event"""
        camera: dict[CAMERA_TYPE, int]

        def __init__(self, rig: R):
            """
            :param rig:``Baselog``
            """
            self.rig = rig

            if rig.version == 'stimpy-git':
                self.camera = {
                    'facecam': 7,
                    'eyecam': 8,
                    '1P_cam': 9,
                }
            else:
                self.camera = {
                    'facecam': 6,
                    'eyecam': 7,
                    '1P_cam': 8,
                }

        def __getitem__(self, item: CAMERA_TYPE) -> CamEvent:
            if item not in self.camera:
                raise KeyError('cam id not found')
            return CamEvent(item, self.rig._event(self.camera[item]))

    @property
    def camera_event(self) -> CameraEvent:
        """camera event. including {'facecam', 'eyecam', '1P_cam'} implemented by __getitem__()"""
        return self.CameraEvent(self)

    # ========================= #
    # Access Other output Files #
    # ========================= #

    @property
    def prot_file(self) -> Path:
        """protocol file path"""
        return self.riglog_file.with_suffix('.prot')

    @abc.abstractmethod
    def get_protocol(self) -> P:
        """
        get protocol (TypeVar ``P``)

        :return: :class:`~stimpyp.base.AbstractStimProtocol()`
        """
        pass

    def get_stimulus_type(self) -> str:
        """get stimulus type name based on protocol"""
        return self.get_protocol().stimulus_type

    @property
    def pref_file(self) -> Path:
        """preferences file path"""
        return self.riglog_file.with_suffix('.prefs')

    def get_preferences(self) -> PreferenceDict:
        """get preferences file"""
        return load_preferences(self.pref_file)

    @abc.abstractmethod
    def get_stimlog(self, *args) -> S:
        """get stimlog (TypeVar ``S``)

        :return: :class:`~AbstractStimlog`
        """
        pass


# =========== #
# Stimlog ABC #
# =========== #

class AbstractStimlog(Generic[R], metaclass=abc.ABCMeta):
    """ABC for stimulation logging. i.e., .log or .stimlog

    `Dimension parameters`:

        N = number of visual stimulation (on-off pairs) = (T * S)

        T = number of trials

        S = number of Stim Type

        P = number of acquisition sample pulse

    """

    # =========== #
    # Log Headers #
    # =========== #
    config = {}
    """config for non-header information"""

    log_info = {}
    """code:log name dictionary"""

    log_header = {}
    """code:log header dictionary"""

    # =========== #
    # Common Attr #
    # =========== #

    time: np.ndarray
    """acquisition time in sec. Array[float, P]"""

    stim_index: np.ndarray
    """stimulation index. Array[int, P]"""

    trial_index: np.ndarray
    """trial index. Array[int, P]"""

    contrast: np.ndarray
    """stimulus contrast. Array[int, P]. value domain in (0,1)"""

    frame_index: np.ndarray
    """stimulation frame index, recount every N. Array[int, P]"""

    # ======= #
    # Grating #
    # ======= #

    ori: np.ndarray
    """direction in deg. Array[float, P]"""

    sf: np.ndarray
    """spatial frequency in cyc/deg. Array[float, P]"""

    tf: np.ndarray
    """temporal frequency in hz. Array[float, P]"""

    phase: np.ndarray
    """stimulus phase for each N. Array[float, P]"""

    # ============= #
    # Function Base #
    # ============= #

    pos_x: np.ndarray
    """object center position X. Array[float, P]"""

    pos_y: np.ndarray
    """object center position Y. Array[float, P]"""

    size_x: np.ndarray
    """object size width. Array[int, P]"""

    size_y: np.ndarray
    """object size height. Array[int, P]"""

    # ====== #
    # Others #
    # ====== #

    flick: np.ndarray
    """TODO. Array[int, P]"""

    interpolate: np.ndarray
    """whether do the interpolate, Array[bool, P]"""

    mask: np.ndarray
    """TODO. Array[bool|None, P]"""

    pattern: np.ndarray
    """object pattern. Array[str, P]"""

    def __init__(self, riglog: R,
                 file_path: PathLike | None,
                 reset_mapping: dict[int, list[str]] | None = None):
        """
        :param riglog: :class:`Baselog`
        :param file_path: filepath of stimlog. could be None if shared log (pyvstim case)
        :param reset_mapping: Customized mapping for ``reset()``
            - key: corresponding to :attr:`log_header`
            - value: list of field, should be the same name as class annotations
        """
        self.riglog_data = riglog
        if file_path is not None:
            self.stimlog_file = Path(file_path)

        self._reset_mapping = reset_mapping

    @abc.abstractmethod
    def _reset(self) -> None:
        """used for assign attributes"""
        pass

    def _reset_cust_mapping(self):
        pass

    # ============ #
    # As Dataframe #
    # ============ #

    @abc.abstractmethod
    def get_visual_stim_dataframe(self, **kwargs) -> pl.DataFrame:
        """Visual presentation dataframe"""
        pass

    @abc.abstractmethod
    def get_state_machine_dataframe(self) -> pl.DataFrame:
        """State Machine dataframe"""
        pass

    def get_photo_indicator_dataframe(self) -> pl.DataFrame:
        """Photo Indicator dataframe. Github version only"""
        pass

    def get_log_dict_dataframe(self) -> pl.DataFrame:
        """Log dict dataframe. Github version only"""
        pass

    # ========= #
    # Time Info #
    # ========= #

    @property
    @abc.abstractmethod
    def exp_start_time(self) -> float:
        """experimental start time (in sec, synced to riglog time with diode offset)"""
        pass

    @property
    @abc.abstractmethod
    def exp_end_time(self) -> float:
        """experimental end time (in sec, synced to riglog time with diode offset)"""
        pass

    @property
    def stim_start_time(self) -> float:
        """the first stimulation start time (in sec, synced to riglog time with diode offset)"""
        return float(self.stimulus_segment[0, 0])

    @property
    def stim_end_time(self) -> float:
        """the last stimulation end time (in sec, synced to riglog time with diode offset)"""
        return float(self.stimulus_segment[-1, 1])

    @property
    @abc.abstractmethod
    def stimulus_segment(self) -> np.ndarray:
        """stimulation time segment (on-off) in sec (N, 2)"""
        pass

    def stim_square_pulse_event(self, sampling_rate: float = 30.) -> RigEvent:
        """
        Get the stimulation on-off square pulse 0,1 consecutive event

        :param sampling_rate: sampling rate for the time domain interpolation
        :return: Stimulus rig event
        """
        start_time = self.exp_start_time
        end_time = self.exp_end_time
        seg = self.stimulus_segment

        t = np.arange(start_time, end_time, 1 / sampling_rate)
        ret = np.zeros_like(t)
        for (on, off) in seg:
            mask = np.logical_and(on < t, t < off)
            ret[mask] = 1

        return RigEvent('visual_stim', np.vstack((t, ret)).T)

    @abc.abstractmethod
    def session_trials(self) -> dict[Session, SessionInfo]:  # TODO might not generic enough
        """get the session:SessionInfo dictionary (experimental and user-specific)"""
        pass

    @property
    @abc.abstractmethod
    def time_offset(self) -> float | np.ndarray:
        """time (in sec) to sync stimlog time to riglog"""
        pass

    # =========================== #
    # Stim Trial/Index/Cycle Info #
    # =========================== #

    @property
    def n_cycles(self) -> list[int]:
        """Number of cycle for each trial"""
        raise [1]

    @property
    @abc.abstractmethod
    def profile_dataframe(self) -> pl.DataFrame:
        """
        Dataframe with columns:

        - stim type index: ``i_stims``
        - trial index: ``i_trials``
        """
        pass

    # ================= #
    # Stim Pattern Info #
    # ================= #

    @abc.abstractmethod
    def get_stim_pattern(self, **kwargs) -> AbstractStimulusPattern:
        """get pattern foreach stimulation"""
        pass


# ================= #
# Protocol File ABC #
# ================= #

class AbstractStimProtocol(metaclass=abc.ABCMeta):
    """ABC class for the stimpy protocol file (.prot)

    `Dimension parameters`:

        N = numbers of visual stimulation (on-off pairs) = (T * S)

        T = number of trials

        S = Number of Stim Type

        C = Numbers of Cycle
    """

    name: str
    """protocol name. related to filename"""

    options: dict[str, Any]
    """protocol options"""

    visual_stimuli_dataframe: pl.DataFrame
    """visual stimuli dataframe. row number: S"""

    version: STIMPY_SOURCE_VERSION
    """date of major changes"""

    def __repr__(self):
        ret = list()

        ret.append('# general parameters')
        for k, v in self.options.items():
            ret.append(f'{k} = {v}')
        ret.append('# stimulus conditions')
        ret.append('\t'.join(self.stim_params))

        ret.append(printdf(self.visual_stimuli_dataframe, do_print=False))

        return '\n'.join(ret)

    def __init__(self, name: str,
                 options: dict[str, Any],
                 visual_stimuli: pl.DataFrame,
                 version: STIMPY_SOURCE_VERSION):

        self.name = name
        self.options = options
        self.visual_stimuli_dataframe = visual_stimuli
        self.version = version

        self._visual_stimuli = None

    @property
    def n_stimuli(self) -> int:
        """number of stimuli (S)"""
        return self.visual_stimuli_dataframe.shape[0]

    @property
    def stim_params(self) -> tuple[str, ...]:
        return tuple(self.visual_stimuli_dataframe.columns)

    @overload
    def __getitem__(self, item: int) -> pl.DataFrame:
        """get row of visual stimuli"""
        pass

    @overload
    def __getitem__(self, item: str) -> np.ndarray:
        """get header of stimuli"""
        pass

    def __getitem__(self, item: str | int) -> np.ndarray | pl.DataFrame:
        """Get protocol value of parameter *item*

        :param item: parameter name. either the row number(int) or field name (str)
        :return: protocol value
        :raises TypeError: `item` not a `str` or `int`
        """
        if isinstance(item, str):
            try:
                return self.visual_stimuli_dataframe.get_column(item).to_numpy()
            except ColumnNotFoundError as e:
                print(f'INVALID: {item}, select from {tuple(self.visual_stimuli_dataframe.columns)}')
                raise e

        elif isinstance(item, int):
            ret = {
                h: self.visual_stimuli_dataframe.row(item)[i]
                for i, h in enumerate(self.visual_stimuli_dataframe.columns)
            }
            return pl.DataFrame(ret)
        else:
            raise TypeError(f'{type(item)}')

    @classmethod
    @abc.abstractmethod
    def load(cls, file: Path | str) -> Self:
        r"""
        Load \*.prot file

        :param file: file path
        """
        pass

    @property
    @abc.abstractmethod
    def is_shuffle(self) -> bool:
        """if shuffle stimulation"""
        pass

    @property
    @abc.abstractmethod
    def background(self) -> float:
        """background for non-stimulation epoch.
        Note the default value need to check in user-specific stimpy version
        """
        pass

    @property
    @abc.abstractmethod
    def start_blank_duration(self) -> int:
        """blank duration before starting the visual stimulation epoch (in sec)"""
        pass

    @property
    @abc.abstractmethod
    def blank_duration(self) -> int:
        """blank duration between each visual stimulus (in sec)"""
        pass

    @property
    @abc.abstractmethod
    def trial_blank_duration(self) -> int:
        """blank duration between trials (in sec)
        TODO check stimpy source code"""
        pass

    @property
    @abc.abstractmethod
    def end_blank_duration(self) -> int:
        """blank duration after the visual stimulation epoch"""
        pass

    @property
    @abc.abstractmethod
    def trial_duration(self) -> int:
        """trial duration"""
        pass

    @property
    def visual_duration(self) -> int:
        """total visual duration"""
        return self.trial_duration * self.n_trials

    @property
    def total_duration(self) -> int:
        """total protocol duration"""
        return self.start_blank_duration + self.visual_duration + self.end_blank_duration

    @property
    def stimulus_type(self) -> str:
        """stimulus type"""
        return self.options['stimulusType']

    @property
    def n_trials(self) -> int:
        """(T,)"""
        return self.options['nTrials']


# ==================== #
# Stimulus Pattern ABC #
# ==================== #


ST = TypeVar('ST')  # Individual Stim


class AbstractStimulusPattern(Generic[ST], metaclass=abc.ABCMeta):
    """
    Abstract Stimulus Pattern

    `Dimension parameters`:

        N = numbers of visual stimulation (on-off pairs) = (T * S)
    """

    time: np.ndarray
    """stim on-off in sec. Array[float, [N, 2]]"""
    contrast: np.ndarray
    """stimulus contrast. Array[float, N]"""
    duration: np.ndarray
    """theoretical duration in prot file, not actual detected using diode. Array[float, N]"""

    def __init__(self, time: np.ndarray,
                 contrast: np.ndarray, *,
                 duration: np.ndarray | None = None):
        """

        :param time: stim on-off in sec. Array[float, [N, 2]]
        :param contrast: stimulus contrast. Array[float, N]
        :param duration: theoretical duration in prot file, not actual detected using diode. Array[float, N]
        """
        self.time = time
        self.contrast = contrast
        self.duration = duration

    @classmethod
    def of(cls, rig: R) -> Self:
        """
        init from Baselog children class

        :param rig: :class:`~stimpyp.base.AbstractLog`
        :return: :class:`StimPattern`
        """
        return rig.get_stimlog().get_stim_pattern()

    @abc.abstractmethod
    def foreach_stimulus(self, name: bool = False) -> Iterable[tuple[Any, ...] | ST]:
        pass



from pathlib import Path

import numpy as np
from matplotlib import pyplot as plt
from matplotlib.axes import Axes

__all__ = [
    'plot_signal_diff',
    'plot_scatter_animation'
]


def plot_signal_diff(t: np.ndarray,
                     alpha: float = 0.6,
                     label: str | None = None,
                     ax: Axes | None = None,
                     **kwargs):
    """
    Plots the reciprocal of the difference of subsequent elements in the input
    array ``t``. This is commonly used for visualizing a derived signal or property
    based on the change between consecutive data points in the input array.
    Additional configurations can be applied to the plot via keyword arguments
    (**kwargs).

    :param t: Input array representing the signal or data points.
        The function computes the reciprocal of the difference between consecutive
        elements.
    :param alpha: Transparency of the plot.
    :param label: Label for the plot.
    :param ax: Matplotlib Axes object to be used for plotting. If None is provided,
        a new Axes instance will be created.
    :param kwargs: Additional keyword arguments for customizing the plot. These
        are passed to the set() method of the Axes object.
    :return: The Axes object containing the plot. If a new Axes instance was
        created internally, it is returned here.
    """
    if ax is None:
        fig, ax = plt.subplots()

    ax.plot(t, 1 / np.diff(t, append=t[0]), alpha=alpha, label=label)
    ax.set(**kwargs)


def plot_scatter_animation(x: np.ndarray,
                           y: np.ndarray,
                           t: np.ndarray | None = None, *,
                           step: int | None = None,
                           size: int = 10,
                           output: Path | str | None = None,
                           **kwargs) -> None:
    r"""
    Plot xy scatter animation with given time points

    :param x: x loc. `Array[float, T]`
    :param y: y loc. `Array[float, T]`
    :param t: time array in sec. `Array[float, T]`
    :param size: size of the scatter
    :param step: step run per datapoint
    :param output: output for animation. i.e., \*.gif
    :param kwargs: additional arguments passed to ``FuncAnimation()``
    :return:
    """
    from matplotlib.animation import FuncAnimation

    fig, _ = plt.subplots()

    def foreach_run(frame: int):
        fig.clear()
        ax = fig.add_subplot(111)
        ax.set_xlim(np.min(x), np.max(x))
        ax.set_ylim(np.min(y), np.max(y))

        if step is not None:
            frame *= step

        ax.text(0.02, 0.95, f'Frames = {frame}', transform=ax.transAxes)
        if t is not None:
            ax.text(0.02, 0.85, f'Time = {t[frame]:.2f}', transform=ax.transAxes)

        ax.scatter(x[frame], y[frame], s=size)

    ani = FuncAnimation(fig, foreach_run, frames=len(x), **kwargs)

    try:
        if output is not None:
            ani.save(output)
        else:
            plt.show()

    finally:
        plt.clf()
        plt.close('all')



from io import BufferedIOBase, BufferedReader
from pathlib import Path
from typing import Union, BinaryIO

__all__ = ['PathLike']

PathLike = Union[str, Path, bytes, BinaryIO, BufferedIOBase, BufferedReader]



import logging

from rich.logging import RichHandler

from .base import *
from .camlog import *
from .event import *
from .log_config import *
from .pyvstim import *
from .session import *
from .stimpy_core import *
from .stimpy_git import *
from .stimulus import *

logger = logging.getLogger("stimpyp")

if not logger.hasHandlers():
    handler = RichHandler()
    formatter = logging.Formatter(
        "%(asctime)s %(levelname)s %(name)s: %(message)s",
        datefmt="[%H:%M:%S]"
    )
    handler.setFormatter(formatter)

    logger.setLevel(logging.INFO)  # default level
    handler.setLevel(logging.INFO)
    logger.addHandler(handler)
    logger.propagate = False



import ast
import logging
import re
from io import StringIO
from pathlib import Path
from typing import Any, final, Literal

import numpy as np
import polars as pl

from ._type import PathLike
from ._util import deprecated_func
from .base import AbstractStimlog
from .session import Session, SessionInfo, get_protocol_sessions
from .stimpy_core import RiglogData, StimpyProtocol
from .stimulus import GratingPattern

__all__ = ['StimlogGit',
           'lazy_load_stimlog']

logger = logging.getLogger(__name__)


# TODO check offset issue with true dataset
@final
class StimlogGit(AbstractStimlog):
    """class for handle the stimlog file for stimpy **github** version
    (mainly tested in the commits derived from master branch)

    `Dimension parameters`:

        N = number of visual stimulation (on-off pairs) = (T * S)

        T = number of trials

        S = number of Stim Type

        I = number of photo indicator pulse
    """

    log_info: dict[int, str]
    """{0: <'Gratings', ...>, 1: 'PhotoIndicator', 2: 'StateMachine', 3: 'LogDict'}. empty if as csv output"""
    log_header: dict[int, list[str]]
    """list of header. empty if as csv output"""
    log_data: dict[int, list[tuple[Any, ...]]]
    """only for StateMachine(2) and LogDict(3)"""

    # =========== #
    # Visual Pars #
    # =========== #

    duration: np.ndarray
    """duration in sec. Array[float, P]"""
    pos_xy: np.ndarray
    """object center position XY. Array[float, [P, 2]]"""
    size_xy: np.ndarray
    """object size width and height. Array[int, [P, 2]]"""

    # ============== #
    # PhotoIndicator #
    # ============== #

    photo_time: np.ndarray
    """photoindicator time. Array[float, I]"""
    photo_state: np.ndarray
    """photoindicator state. Array[bool, I]"""
    photo_size: np.ndarray
    """photoindicator size in given unit. Array[float, I]"""
    photo_pos: np.ndarray
    """photoindicator in XY. Array[int, [I, 2]]"""
    photo_units: np.ndarray
    """photoindicator unit. Array[str, I]"""
    photo_mode: np.ndarray
    """photoindicator mode. Array[int, I]"""
    photo_frames: np.ndarray
    """photoindicator frames . Array[int, I]"""
    photo_enable: np.ndarray
    """photoindicator size in pixel. Array[bool, I]"""

    def __init__(self, riglog: RiglogData,
                 file_path: PathLike,
                 reset_mapping: dict[int, list[str]] | None = None,
                 csv_output: bool = False,
                 diode_offset: bool = True):
        """
        init based on :class:`~stimpyp.stimpy_core.RiglogData` information

        :param riglog: :class:`~stimpyp.stimpy_core.RiglogData`
        :param file_path: filepath of stimlog. could be None if shared log (pyvstim case)
        :param reset_mapping: Customized mapping for ``reset()``
            - key: corresponding to :attr:`log_header`
            - value: list of field, should be the same name as class annotations
        :param csv_output: if stimlog is exported to separated csv file
        :param diode_offset: do the offset to sync time with :class:`~stimpyp.stimpy_core.RiglogData`
        """

        super().__init__(riglog, file_path, reset_mapping)

        self.diode_offset = diode_offset

        self._csv_output = csv_output
        self._csv_added = []

        self._cache_time_offset: float | None = None
        self._cache_visual_stim_dataframe: pl.DataFrame | None = None
        self._cache_photo_indicator_dataframe: pl.DataFrame | None = None
        self._cache_state_machine_dataframe: pl.DataFrame | None = None
        self._cache_log_dict_dataframe: pl.DataFrame | None = None

        self._reset()

    def _reset(self):
        log_data = {}

        with self.stimlog_file.open() as f:
            for line, content in enumerate(f):  # type: int, str
                content = content.strip()
                # `## (CODE VERSION) : (commit hash: 88c4705 - tags: [''])`
                # *: 0-inf; +: 1-inf; ?: 0-1
                m = re.match(r'#+ (.+?)\s*:\s*(.+)', content)
                if m:
                    info_name = m.group(1)
                    info_value = m.group(2)

                    if info_name == 'LOG NAME':
                        self.config['log_name'] = info_value
                    elif info_name == 'CODE VERSION':
                        self._reset_code_version(info_value)
                    elif info_name == 'Format':
                        self.config['format'] = info_value.split(' ')
                        if self.config['format'] != ['source_id', 'time', 'source_infos']:
                            raise RuntimeError('stimlog format changed')
                    elif info_name == 'Rig trigger on':
                        info_value = info_value.split(',')
                        self.config['rig_trigger'] = (info_value[0], float(info_value[1]))
                    elif self._reset_log_info(info_name, info_value):
                        pass
                    else:
                        logger.info(f'ignore header line at {line + 1} : {content}')

                elif content.startswith('### START'):
                    self.config['start_time'] = float(content[9:].strip())
                elif content.startswith('### END'):
                    self.config['end_time'] = float(content[7:].strip())
                elif content.startswith('# Missed') and content.endswith('frames'):
                    self.config['missed_frames'] = int(content.split(' ')[2])
                    logger.info(f'missed frames: {self.config["missed_frames"]}')
                elif self._csv_output and content.startswith('#') and content.endswith('added'):  # csv
                    self._csv_added.append(content.split(' ')[1])
                elif content.startswith('#'):  # other
                    logger.info(f'ignore header line at {line + 1} : {content}')
                else:
                    if not self._csv_output:
                        try:
                            self._reset_line(log_data, line + 1, content)
                        except BaseException:
                            raise RuntimeError(f'bad format line at {line + 1} : {content}')

        if self._csv_output:
            logger.info(f'stimlog CSV output: {self._csv_added}')
            for log_type in self._csv_added:
                self._reset_csv(log_type)

        self.log_data = self._reset_final(log_data)

    def _reset_code_version(self, info_value: str):
        """for ### CODE VERSION, commit hash and tags"""
        info_value = info_value.strip().split(' ')
        commits = info_value[info_value.index('hash:') + 1]
        tag = info_value[info_value.index('tags:') + 1]

        self.config['commit_hash'] = commits if commits != 'None' else None
        self.config['tag'] = tag if tag != 'None' else None

    def _reset_log_info(self, code: str, info_value: str):
        try:
            code = int(code)
        except ValueError:
            return False

        name, header = info_value.split(' ', maxsplit=1)
        header = eval(header)
        self.log_info[code] = name
        self.log_header[code] = header
        return True

    # ======================== #
    # Reset All-in-One Stimlog #
    # ======================== #

    def _reset_line(self, log_data: dict[int, list], line: int, content: str):
        message: str
        value: list
        code, time, message = content.split(' ', maxsplit=2)
        code = int(code)
        time = float(time)

        if self.log_info[code] in ('Gratings', 'FunctionBased', 'PhotoIndicator', 'LogDict'):
            value = eval(message)
            if len(self.log_header[code]) != len(value):
                logger.warning(f'log category {code} size mismatched at line {line} : {message}')
            else:
                log_data.setdefault(code, []).append((time, *value))

        elif self.log_info[code] == 'StateMachine':
            value = eval(message.replace('<', '("').replace(':', '",').replace('>', ')'))
            value = list(map(str, value))
            if len(self.log_header[code]) != len(value):
                logger.warning(f'log category {code} size mismatched at line {line} : {message}')
            else:
                log_data.setdefault(code, []).append((time, *value))

        else:
            logger.warning(f'unknown log category : at line {line} : {content}')

    def _reset_final(self, log_data: dict[int, list]) -> dict[int, list]:
        """as attr"""
        remove_code = []
        for code, content in log_data.items():
            if self.log_info[code] in ('Gratings', 'FunctionBased'):
                self.time = np.array([it[0] for it in content])
                self.duration = np.array([it[1] for it in content])
                self.contrast = np.array([it[2] for it in content])
                self.ori = np.array([it[3] for it in content])
                self.phase = np.array([it[4] for it in content])
                self.pos_xy = np.array([it[5] for it in content])
                self.size_xy = np.array([it[6] for it in content])
                self.flick = np.array([it[7] for it in content])
                self.interpolate = np.array([it[8] for it in content], dtype=bool)
                self.mask = np.array([it[9] is not None for it in content], dtype=bool)
                self.sf = np.array([it[10] for it in content])
                self.tf = np.array([it[11] for it in content])
                self.opto = np.array([it[12] for it in content], dtype=int)
                self.pattern = np.array([it[13] for it in content])

                remove_code.append(code)

            elif self.log_info[code] == 'PhotoIndicator':
                self.photo_time = np.array([it[0] for it in content])
                self.photo_state = np.array([it[1] for it in content], dtype=bool)
                self.photo_size = np.array([it[2] for it in content])
                self.photo_pos = np.array([it[3] for it in content])
                self.photo_units = np.array([it[4] for it in content])
                self.photo_mode = np.array([it[5] for it in content], dtype=int)
                self.photo_frames = np.array([it[6] for it in content], dtype=int)
                self.photo_enable = np.array([it[7] for it in content], dtype=bool)

                remove_code.append(code)

        for code in remove_code:
            del log_data[code]

        return log_data

    # =================== #
    # Reset Separated CSV #
    # =================== #

    def _reset_csv(self, log_type: Literal['FunctionBased', 'Gratings', 'LogDict', 'PhotoIndicator', 'StateMachine']):
        file = self.stimlog_file.parent / f'{self.stimlog_file.stem}_{log_type}.csv'

        if 'Gratings' in log_type:
            self._reset_grating_csv(file)
        elif 'PhotoIndicator' in log_type:
            self._reset_photo_indicator_csv(file)
        elif 'StateMachine' in log_type:
            self._reset_state_machine_csv(file)
        elif 'LogDict' in log_type:
            self._reset_log_dict_csv(file)
        else:
            raise NotImplementedError(f'{log_type} not yet supported')

    def _reset_grating_csv(self, file: Path, cleanup: bool = True):
        """

        :param file:
        :param cleanup: remove redundant datapoint in ``time`` column
        :return:
        """
        df = list_value_parser(file, cols=['pos', 'size'])

        if cleanup:  # remove redundant datapoint
            orig = df.columns
            cols = [col for col in orig if col != 'time']

            df = (
                df.group_by(cols, maintain_order=True)
                .agg(pl.col('*').first())
                .sort('time')
                .select(orig)  # restore original column order
            )

        self._cache_visual_stim_dataframe = df
        self.time = df['time'].to_numpy()
        self.duration = df['duration'].to_numpy()
        self.contrast = df['contrast'].to_numpy()
        self.ori = df['ori'].to_numpy()
        self.phase = df['phase'].to_numpy()
        self.pos_xy = df['pos'].to_numpy()
        self.size_xy = df['size'].to_numpy()
        self.flick = df['flick'].to_numpy()
        self.interpolate = df['interpolate'].to_numpy()
        self.mask = df['mask'].to_numpy()
        self.sf = df['sf'].to_numpy()
        self.tf = df['tf'].to_numpy()
        self.opto = df['opto'].to_numpy()
        self.pattern = df['pattern'].to_numpy()

    def _reset_photo_indicator_csv(self, file: Path, cleanup: bool = True):
        """

        :param file:
        :param cleanup: remove redundant datapoint keep first for photodiode ``state`` switching
        :return:
        """
        df = list_value_parser(file, cols=['pos'])

        if cleanup:
            df = df.filter(
                (pl.col("state") != pl.col("state").shift(1)) |
                (pl.arange(0, df.height) == 0)  # keep first row
            )

        self._cache_photo_indicator_dataframe = df
        self.photo_time = df['time'].to_numpy()
        self.photo_state = df['state'].to_numpy()
        self.photo_size = df['size'].to_numpy()
        self.photo_pos = df['pos'].to_numpy()
        self.photo_units = df['units'].to_numpy()
        self.photo_mode = df['mode'].to_numpy()
        self.photo_frames = df['frames'].to_numpy()
        self.photo_enable = df['enabled'].to_numpy()

    def _reset_state_machine_csv(self, file: Path, cleanup: bool = True):
        """

        :param file:
        :param cleanup: remove redundant datapoint keep first for statemachine ``state`` switching
        :return:
        """
        df = pl.read_csv(file)

        if cleanup:
            df = df.filter((pl.col("state") != pl.col("state").shift(1)) | (pl.arange(0, df.height) == 0))

        self._cache_state_machine_dataframe = df

    def _reset_log_dict_csv(self, file: Path, cleanup: bool = True):
        """

        :param file:
        :param cleanup: remove redundant datapoint in ``time``
        :return:
        """
        df = pl.read_csv(file)

        if cleanup:
            orig = df.columns
            cols = [col for col in orig if col != 'time']

            df = (
                df.group_by(cols, maintain_order=True)
                .agg(pl.col('*').first())
                .sort('time')
                .select(orig)  # restore original column order
            )

        self._cache_log_dict_dataframe = df

    # ============= #
    # Get DataFrame #
    # ============= #

    def get_visual_stim_dataframe(self) -> pl.DataFrame:
        """
        Visual presentation dataframe::

            ┌────────────┬──────────┬──────────┬─────┬───┬──────┬─────┬──────┬─────────┐
            │ time       ┆ duration ┆ contrast ┆ ori ┆ … ┆ sf   ┆ tf  ┆ opto ┆ pattern │
            │ ---        ┆ ---      ┆ ---      ┆ --- ┆   ┆ ---  ┆ --- ┆ ---  ┆ ---     │
            │ f64        ┆ i64      ┆ i64      ┆ i64 ┆   ┆ f64  ┆ i64 ┆ i64  ┆ str     │
            ╞════════════╪══════════╪══════════╪═════╪═══╪══════╪═════╪══════╪═════════╡
            │ 18.990026  ┆ 10       ┆ 1        ┆ 0   ┆ … ┆ 0.04 ┆ 10  ┆ 0    ┆ square  │
            │ 21.000029  ┆ 10       ┆ 1        ┆ 0   ┆ … ┆ 0.04 ┆ 10  ┆ 0    ┆ square  │
            │ …          ┆ …        ┆ …        ┆ …   ┆ … ┆ …    ┆ …   ┆ …    ┆ …       │
            │ 619.054972 ┆ 10       ┆ 1        ┆ 0   ┆ … ┆ 0.04 ┆ 50  ┆ 0    ┆ square  │
            │ 619.084972 ┆ 10       ┆ 1        ┆ 0   ┆ … ┆ 0.04 ┆ 50  ┆ 0    ┆ square  │
            └────────────┴──────────┴──────────┴─────┴───┴──────┴─────┴──────┴─────────┘

        :return:
        """
        if self._cache_visual_stim_dataframe is None:
            self._cache_visual_stim_dataframe = pl.DataFrame().with_columns(
                pl.Series(self.time).alias('time'),
                pl.Series(self.duration).alias('duration'),
                pl.Series(self.contrast).alias('contrast'),
                pl.Series(self.ori).alias('ori'),
                pl.Series(self.phase).alias('phase'),
                pl.Series(self.pos_xy).alias('pos'),
                pl.Series(self.size_xy).alias('size'),
                pl.Series(self.flick).alias('flick'),
                pl.Series(self.interpolate).alias('interpolate'),
                pl.Series(self.mask).alias('mask'),
                pl.Series(self.sf).alias('sf'),
                pl.Series(self.tf).alias('tf'),
                pl.Series(self.opto).alias('opto'),
                pl.Series(self.pattern).alias('pattern')
            )
        return self._cache_visual_stim_dataframe

    def get_photo_indicator_dataframe(self) -> pl.DataFrame:
        """
        PhotoDiode dataframe::

            ┌────────────┬───────┬──────┬────────────┬───────┬──────┬────────┬────────┐
            │ time       ┆ state ┆ size ┆ pos        ┆ units ┆ mode ┆ frames ┆ enable │
            │ ---        ┆ ---   ┆ ---  ┆ ---        ┆ ---   ┆ ---  ┆ ---    ┆ ---    │
            │ f64        ┆ bool  ┆ i64  ┆ list[i64]  ┆ str   ┆ i64  ┆ i64    ┆ bool   │
            ╞════════════╪═══════╪══════╪════════════╪═══════╪══════╪════════╪════════╡
            │ 18.990026  ┆ false ┆ 60   ┆ [740, 370] ┆ pix   ┆ 0    ┆ 20     ┆ true   │
            │ 21.000029  ┆ true  ┆ 60   ┆ [740, 370] ┆ pix   ┆ 0    ┆ 20     ┆ true   │
            │ …          ┆ …     ┆ …    ┆ …          ┆ …     ┆ …    ┆ …      ┆ …      │
            │ 607.094955 ┆ false ┆ 60   ┆ [740, 370] ┆ pix   ┆ 0    ┆ 20     ┆ true   │
            │ 609.104958 ┆ true  ┆ 60   ┆ [740, 370] ┆ pix   ┆ 0    ┆ 20     ┆ true   │
            └────────────┴───────┴──────┴────────────┴───────┴──────┴────────┴────────┘

        :return:
        """
        if self._cache_photo_indicator_dataframe is None:
            self._cache_photo_indicator_dataframe = pl.DataFrame().with_columns(
                pl.Series(self.photo_time).alias('time'),
                pl.Series(self.photo_state).alias('state'),
                pl.Series(self.photo_size).alias('size'),
                pl.Series(self.photo_pos).alias('pos'),
                pl.Series(self.photo_units).alias('units'),
                pl.Series(self.photo_mode).alias('mode'),
                pl.Series(self.photo_frames).alias('frames'),
                pl.Series(self.photo_enable).alias('enable')
            )
        return self._cache_photo_indicator_dataframe

    def get_state_machine_dataframe(self) -> pl.DataFrame:
        """
        State Machine dataframe::

            ┌────────────┬───────────────────────────┬───────────────────────────┐
            │ time       ┆ state                     ┆ prev_state                │
            │ ---        ┆ ---                       ┆ ---                       │
            │ f64        ┆ str                       ┆ str                       │
            ╞════════════╪═══════════════════════════╪═══════════════════════════╡
            │ 18.990026  ┆ ('States.SHOW_BLANK', 1)  ┆ ('States.STIM_SELECT', 0) │
            │ 20.990029  ┆ ('States.SHOW_STIM', 2)   ┆ ('States.SHOW_BLANK', 1)  │
            │ …          ┆ …                         ┆ …                         │
            │ 609.094958 ┆ ('States.SHOW_STIM', 2)   ┆ ('States.SHOW_BLANK', 1)  │
            │ 619.094972 ┆ ('States.STIM_SELECT', 0) ┆ ('States.SHOW_STIM', 2)   │
            └────────────┴───────────────────────────┴───────────────────────────┘

        :return:
        """
        if self._cache_state_machine_dataframe is None:
            self._cache_state_machine_dataframe = pl.DataFrame().with_columns(
                pl.Series(it[0] for it in self.log_data[2]).alias('time'),
                pl.Series(it[1] for it in self.log_data[2]).alias('state'),
                pl.Series(it[2] for it in self.log_data[2]).alias('prev_state'),
            )
        return self._cache_state_machine_dataframe

    def get_log_dict_dataframe(self) -> pl.DataFrame:
        """
        Log Dict DataFrame::

            ┌────────────┬──────────┬──────────┬──────────────┬────────────┐
            │ time       ┆ block_nr ┆ trial_nr ┆ condition_nr ┆ trial_type │
            │ ---        ┆ ---      ┆ ---      ┆ ---          ┆ ---        │
            │ f64        ┆ i64      ┆ i64      ┆ i64          ┆ i64        │
            ╞════════════╪══════════╪══════════╪══════════════╪════════════╡
            │ 18.990026  ┆ 0        ┆ 0        ┆ 0            ┆ 1          │
            │ 30.990043  ┆ 1        ┆ 0        ┆ 0            ┆ 1          │
            │ …          ┆ …        ┆ …        ┆ …            ┆ …          │
            │ 595.083939 ┆ 48       ┆ 0        ┆ 0            ┆ 1          │
            │ 607.094955 ┆ 49       ┆ 0        ┆ 0            ┆ 1          │
            └────────────┴──────────┴──────────┴──────────────┴────────────┘

        :return:
        """
        if self._cache_log_dict_dataframe is None:
            self._cache_log_dict_dataframe = pl.DataFrame().with_columns(
                pl.Series(it[0] for it in self.log_data[3]).alias('time'),
                pl.Series(it[1] for it in self.log_data[3]).alias('block_nr'),
                pl.Series(it[2] for it in self.log_data[3]).alias('trial_nr'),
                pl.Series(it[3] for it in self.log_data[3]).alias('condition_nr'),
                pl.Series(it[4] for it in self.log_data[3]).alias('trial_type'),
            )

        return self._cache_log_dict_dataframe

    @property
    def pos_x(self) -> np.ndarray:
        """object center position X. Array[float, P]"""
        return self.pos_xy[:, 0]

    @property
    def pos_y(self) -> np.ndarray:
        """object center position Y. Array[float, P]"""
        return self.pos_xy[:, 1]

    @property
    def size_x(self) -> np.ndarray:
        """object size width. Array[int, P]"""
        return self.size_xy[:, 0]

    @property
    def size_y(self) -> np.ndarray:
        """object size height. Array[int, P]"""
        return self.size_xy[:, 1]

    @property
    def exp_start_time(self) -> float:
        return self.riglog_data.dat[0, 2] / 1000

    @property
    def exp_end_time(self) -> float:
        return self.riglog_data.dat[-1, 2] / 1000

    @property
    def stim_start_time(self) -> float:
        return float(self.photo_time[1] + self.time_offset)

    @property
    def stim_end_time(self) -> float:
        return float(self.time[-1] + self.time_offset)

    @property
    def stimulus_segment(self) -> np.ndarray:
        # ret = self.riglog_data.screen_event[:, 0].reshape(-1, 2)  # directly use riglog time
        # stimlog bug starting from `diode off`, and ending from `diode True`
        # thus remove first point and add last point manually
        _p_time = np.concatenate((self.photo_time[1:], np.array([self.time[-1]])))
        ret = _p_time.reshape(-1, 2) + self.time_offset
        return ret

    def session_trials(self) -> dict[Session, SessionInfo]:
        return {
            prot.name: prot
            for prot in get_protocol_sessions(self)
        }

    @property
    def time_offset(self) -> float:
        if self._cache_time_offset is None:
            # noinspection PyTypeChecker
            self._cache_time_offset = _time_offset(self.riglog_data, self, self.diode_offset)
        return self._cache_time_offset

    def get_stim_pattern(self) -> GratingPattern:
        prot = StimpyProtocol.load(self.stimlog_file.with_suffix('.prot'))
        log_nr = self.get_log_dict_dataframe()['condition_nr'].to_numpy()

        t = self.stimulus_segment
        ori = np.array([prot['ori'][n] for n in log_nr])
        sf = np.array([prot['sf'][n] for n in log_nr])
        tf = np.array([prot['tf'][n] for n in log_nr])
        contrast = np.array([prot['c'][n] for n in log_nr])
        dur = np.array([prot['dur'][n] for n in log_nr])

        return GratingPattern(t, ori, sf, tf, contrast, duration=dur)

    @property
    def profile_dataframe(self) -> pl.DataFrame:
        return (
            self.get_log_dict_dataframe()
            .with_columns(pl.col('condition_nr').alias('i_stims'))
            .with_columns(pl.col('block_nr').alias('i_trials'))
            .select('i_stims', 'i_trials')
        )


# ========== #

@deprecated_func(remarks='generalized, to sequential offset method')
def _time_offset(rig: RiglogData,
                 stm: StimlogGit,
                 diode_offset=True) -> float:
    """
    time offset used in the `new stimpy`
    offset time from screen_time .riglog (diode) to .stimlog
    ** normally stimlog time value are larger than riglog

    :param rig:
    :param stm:
    :param diode_offset: whether correct diode signal
    :return: tuple of offset average and std value
    """
    if not isinstance(stm, StimlogGit):
        raise TypeError('')

    screen_time = rig.screen_event.time

    if diode_offset:
        try:
            # new stimpy might be a negative value (stimlog time value larger than riglog)
            # stimlog bug starting from `diode off`, and ending from `diode True`
            # thus remove first point and add last point manually
            _p_time = np.concatenate((stm.photo_time[1:], np.array([stm.time[-1]])))
            offset_t = screen_time[::2] - _p_time[::2]
        except ValueError as e:
            print(f'number of diode pulse and stimulus mismatch from {e}')
            print(f'use the first pulse diff for alignment')
            offset_t = screen_time[0] - stm.photo_time[1]
    else:
        raise NotImplementedError('')

    offset_t_avg = float(np.mean(offset_t))
    offset_t_std = float(np.std(offset_t))

    print(f'time offset between stimlog and riglog: {round(offset_t_avg, 3)}')
    print(f'offset_std: {round(offset_t_std, 3)}')

    return offset_t_avg


def lazy_load_stimlog(file: PathLike, string_key: bool = True) -> dict[str | int, pl.DataFrame]:
    """
    Load directly the stimlog file (without riglog time offset), and parse data as polars dataframes

    .. code-block:: python

        file = ...

        log = load_stimlog(file, string_key=True)  # get dataframe using keyname
        print(log['PhotoIndicator'])

        log = load_stimlog(file, string_key=False)  # get dataframe using code int
        print(log[1])

    :param file: file path for the .stimlog
    :param string_key: show key as str type, otherwise, int type
    :return: Code:DataFrame dictionary
    """

    log_info = {}
    log_header = {}
    log_data = {}

    with Path(file).open() as f:
        for line, content in enumerate(f):
            content = content.strip()

            #
            m = re.match(r'#+ (.+?)\s*:\s*(.+)', content)
            if m:
                info_name = m.group(1)
                info_value = m.group(2)
                try:
                    code = int(info_name)
                except ValueError:
                    continue
                else:
                    name, header = info_value.split(' ', maxsplit=1)
                    header = eval(header)
                    log_info[code] = name
                    log_header[code] = header

            elif content.startswith('#'):
                logger.info(f'ignore header line at {line + 1} : {content}')
                continue

            else:  # data
                code, time, message = content.split(' ', maxsplit=2)
                code = int(code)
                time = float(time)

                if log_info[code] == 'StateMachine':
                    value = eval(message.replace('<', '("').replace(':', '",').replace('>', ')'))
                    value = list(map(str, value))
                    if len(log_header[code]) != len(value):
                        logger.warning(f'log category {code} size mismatched at line {line} : {message}')
                    else:
                        log_data.setdefault(code, []).append((time, *value))

                else:
                    value = eval(message)
                    if len(log_header[code]) != len(value):
                        logger.warning(f'log category {code} size mismatched at line {line} : {message}')
                    else:
                        log_data.setdefault(code, []).append((time, *value))

        return {
            (log_info[code] if string_key else code):
                pl.DataFrame(log_data[code], schema=['time'] + log_header[code], strict=False)
            for code in log_data
        }


def list_value_parser(f: Path, cols: list[str]) -> pl.DataFrame:
    """parse list value in csv file, and return a new dataframe

    .. code-block:: text

        1,2,3,[100,100],5

    **TO**

    .. code-block:: text

        1,2,3,"[100,100]",5


    """

    def list_caster(x, dtype: type = int):
        try:
            return [dtype(i) for i in ast.literal_eval(x)]
        except Exception:
            return x

    raw = f.read_text()
    fixed = re.sub(r'(\[.*?\])', r'"\1"', raw)
    return pl.read_csv(StringIO(fixed)).with_columns(pl.col(c).map_elements(list_caster) for c in cols)



from __future__ import annotations

import logging
import re
import warnings
from pathlib import Path
from typing import Any, final, Iterable

import numpy as np
import polars as pl
from typing_extensions import Self

from ._type import PathLike
from ._util import unfold_stimuli_condition, try_casting_number, cls_hasattr
from .base import AbstractLog, AbstractStimlog, AbstractStimProtocol, AbstractStimulusPattern
from .session import Session, SessionInfo, get_protocol_sessions
from .stimulus import GratingPattern, FunctionPattern

__all__ = [
    'load_riglog',
    'RiglogData',
    #
    'Stimlog',
    #
    'load_protocol',
    'StimpyProtocol'
]

logger = logging.getLogger(__name__)


def load_riglog(root_path: PathLike,
                diode_offset: bool = True,
                reset_mapping: dict[int, list[str]] | None = None) -> RiglogData:
    """
    load riglog data

    :param root_path: riglog file path or riglog directory path
    :param diode_offset: do the diode offset to sync the time between riglog and stimlog
    :param reset_mapping:
    :return: :class:`RiglogData`
    """
    return RiglogData(root_path=root_path, diode_offset=diode_offset, reset_mapping=reset_mapping)


@final
class RiglogData(AbstractLog):
    """class for handle the riglog file for stimpy **bitbucket/github** version
    (mainly tested in the commits derived from master branch)
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, log_suffix='.riglog', **kwargs)

        self.__stimlog_cache: AbstractStimlog | None = None
        self.__prot_cache: StimpyProtocol | None = None

    @classmethod
    def _cache_asarray(cls, filepath: Path, square_brackets: bool) -> np.ndarray:
        output = filepath.with_name(filepath.stem + '_riglog.npy')

        if not output.exists():
            if square_brackets:
                riglog = np.loadtxt(
                    filepath,
                    delimiter=',',
                    comments=['#', 'None'],
                    converters={
                        0: lambda it: float(it[1:]),
                        3: lambda it: float(it[:-1])
                    },
                )
                logger.debug('square bracket were found in riglog file')
            else:
                riglog = pl.read_csv(filepath, comment_prefix='#').to_numpy()

            np.save(output, riglog)

        return np.load(output)

    def with_sessions(self, session: Session | tuple[Session, ...]) -> Self:
        """
        Truncate the instance `dat` with the given session(s)

        :param session: session name(s)
        :return:
        """
        dy = self.get_stimlog().session_trials()

        if isinstance(session, str):
            t0, t1 = dy[session].time
        elif isinstance(session, tuple):
            t_all = np.array([dy[s].time for s in session])
            t0, t1 = np.min(t_all), np.max(t_all)
            logger.info(f'get trange from multiple sessions, {session}: t0:{t0}, t1:{t1}')
        else:
            raise TypeError('')

        t = self.dat[:, 2] / 1000
        mask = np.logical_and(t0 < t, t < t1)
        self.dat = self.dat[mask]

        return self

    @property
    def stimlog_file(self) -> Path:
        return self.riglog_file.with_suffix('.stimlog')

    def get_stimlog(self, csv_output: bool = True) -> AbstractStimlog:
        """
        Initialize the stimlog instance

        :param csv_output: if stimlog is exported to separated csv file
        """
        match self.__stimlog_cache, self.version:
            case (None, 'stimpy-git'):
                from .stimpy_git import StimlogGit
                self.__stimlog_cache = StimlogGit(
                    self,
                    self.stimlog_file,
                    self._reset_mapping,
                    csv_output=csv_output,
                    diode_offset=self._diode_offset
                )
            case (None, 'stimpy-bit'):
                self.__stimlog_cache = Stimlog(self, self.stimlog_file, self._reset_mapping, self._diode_offset)
            case (None, _):
                raise ValueError(f'unknown version: {self.version}')

        logger.debug(f'init stimlog with {type(self.__stimlog_cache).__name__}')
        return self.__stimlog_cache

    def get_protocol(self) -> 'StimpyProtocol':
        if self.__prot_cache is None:
            self.__prot_cache = StimpyProtocol.load(self.prot_file)

        return self.__prot_cache


@final
class Stimlog(AbstractStimlog):
    """class for handle the stimlog file for stimpy **bitbucket** version
    (mainly tested in the commits derived from master branch)

    `Dimension parameters`:

        N = numbers of visual stimulation (on-off pairs) = (T * S)

        T = number of trials

        S = number of Stim Type

        P = number of acquisition sample pulse (Visual parameters)

        M = number of statemachine pulse
    """

    config: dict[str, Any] = {}
    """i.e., name, commit hash, missed_frames, ..."""

    log_info: dict[int, str] = {}
    """i.e., {10: 'vstim', 20: 'stateMachine'}"""

    log_header: dict[int, list[str]] = {}
    """i.e., {10: ['code','presentTime','iStim', ...], 20: ['code', 'elapsed', 'cycle', ...]}"""

    photo_state: np.ndarray
    """photo diode on-off. Array[int, P]. value domain in (0,1)"""

    indicator_flag: np.ndarray
    """accumulated and zero foreach stim. Array[int, P]. value domain in (1, ...)"""

    # ========================================= #
    # StateMachine logging (start from code 20) #
    # ========================================= #

    s_on_v: np.ndarray
    """statemachine sync time on v, use for time sync if diode signal not reliable. Array[float, M]"""

    state_time: np.ndarray
    """Array[float, M]"""

    state_cycle: np.ndarray
    """Array[float, M]"""

    state_new_state: np.ndarray
    """Array[float, M]"""

    state_old_state: np.ndarray
    """Array[float, M]"""

    state_elapsed: np.ndarray
    """Array[float, M]"""

    state_trial_type: np.ndarray
    """Array[float, M]"""

    def __init__(self,
                 riglog: RiglogData,
                 file_path: PathLike,
                 reset_mapping: dict[int, list[str]] | None = None,
                 diode_offset: bool = True,
                 sequential_offset: bool = True):
        """
        :param riglog: ``RiglogData``
        :param file_path: stimlog filepath
        :param diode_offset: If do the diode offset to sync the time to riglog
        :param sequential_offset: do the sequential offset to sync time with :class:`RiglogData`
        """

        super().__init__(riglog, file_path, reset_mapping)
        self.diode_offset = diode_offset

        # diode
        self._do_sequential_diode_offset = sequential_offset
        self._cache_time_offset: float | np.ndarray | None = None

        self._reset()

    def _reset(self):
        try:
            stim_type = self.riglog_data.get_stimulus_type()
        except AttributeError:
            print(f'no riglog init, for only testing, some methods might causes problem', vtype='warning')
            return self._reset_gratings()  # testing

        match stim_type, self._reset_mapping:
            case ('gratings', None):
                self._reset_gratings()
                return None
            case ('functions', None):
                self._reset_functions()
                return None
            case (_, mapping) if mapping is not None:
                self._reset_cust_mapping()
                return None
            case _:
                raise NotImplementedError('')

    def _eager_comment_info(self):
        """Eager get the #comment_info"""
        with self.stimlog_file.open() as _f:
            for number, line in enumerate(_f):
                line = line.strip()

                if len(line) == 0:
                    continue

                if line.startswith('#'):
                    self._reset_comment_info(line)

    def _reset_cust_mapping(self):
        def _cast(s) -> int | float:
            try:
                v = float(s)
                if v.is_integer():
                    return int(v)
                else:
                    return v
            except (ValueError, IndexError):
                return -1

        self._eager_comment_info()
        vlog = [[] for _ in range(len(self.log_header[10]) - 1)]  # without code
        state = [[] for _ in range(len(self.log_header[20]) - 1)]  # without code

        with self.stimlog_file.open() as _f:
            for number, line in enumerate(_f):
                line = line.strip()

                if len(line) == 0 or line.startswith('#'):
                    continue

                part = line.split(',')
                try:
                    code = int(part[0])
                    if code == 10:
                        for i in range(len(vlog)):
                            try:
                                vlog[i].append(_cast(part[i + 1]))
                            except IndexError:
                                vlog[i].append(-1)
                    elif code == 20:
                        for i in range(len(state)):
                            state[i].append(_cast(part[i + 1]))
                    else:
                        raise ValueError(f'unknown code: {code}')

                except BaseException as e:
                    raise RuntimeError(f'error parsing line: {line} in line:{number}, {repr(e)}')

        #
        for code, fields in self._reset_mapping.items():
            if code == 10:
                for i, f in enumerate(fields):
                    setattr(self, f, np.array(vlog[i]))
            elif code == 20:
                for i, f in enumerate(fields):
                    v = np.array(state[i])
                    if f in ('state_time', 'state_elapsed'):  # to sec
                        v = v / 1000
                    setattr(self, f, v)
            else:
                raise ValueError(f'unknown code: {code}')

        self._check_mapping()

    def _check_mapping(self) -> None:
        for code, fields in self._reset_mapping.items():
            # length check
            if len(fields) != len(self.log_header[code]) - 1:
                raise ValueError(f'reset mapping number mismatch: should follow the order: {self.log_header[code]}')

            # name check
            for f in fields:
                if not cls_hasattr(Stimlog, f):
                    raise ValueError(f'reset mapping need to be the one of the annotations: "{f}"')

    def _reset_gratings(self):
        time = []
        stim_index = []
        trial_index = []
        photo_state = []
        contrast = []
        ori = []
        sf = []
        phase = []
        frame_index = []

        s_on_v = []
        s_present_time = []
        s_cycle = []
        s_new_state = []
        s_old_state = []
        s_state_elapsed = []
        s_trial_type = []

        with self.stimlog_file.open() as _f:
            for number, line in enumerate(_f):
                line = line.strip()
                if len(line) == 0:
                    continue

                if line.startswith('#'):
                    self._reset_comment_info(line)
                    continue

                part = line.split(',')

                try:
                    code = int(part[0])

                    if code == 10:
                        time.append(float(part[1]))

                        try:
                            stim_value = int(part[2])
                        except ValueError:  # 'None'
                            stim_value = -1
                        stim_index.append(stim_value)

                        trial_index.append(int(part[3]))
                        photo_state.append(int(part[4]))

                        if len(part) > 5:
                            contrast.append(float(part[5]))
                            ori.append(int(float(part[6])))
                            sf.append(float(part[7]))
                            phase.append(float(part[8]))
                            frame_index.append(int(part[9]))
                        else:
                            # empty list item = -1
                            contrast.append(-1)
                            ori.append(-1)
                            sf.append(-1)
                            phase.append(-1)
                            frame_index.append(-1)

                    elif code == 20:
                        s_on_v.append(time[-1])  # sync s to v
                        s_present_time.append(int(part[1]) / 1000)
                        s_cycle.append(int(part[2]))
                        s_new_state.append(int(part[3]))
                        s_old_state.append(int(part[4]))
                        s_state_elapsed.append(int(part[5]) / 1000)
                        s_trial_type.append(int(part[6]))

                    else:
                        raise ValueError(f'unknown code : {code}')

                except BaseException as e:
                    raise RuntimeError(f'line {number}: {line}') from e

        self.time = np.array(time)
        self.stim_index = np.array(stim_index)
        self.trial_index = np.array(trial_index)
        self.photo_state = np.array(photo_state)
        self.contrast = np.array(contrast)
        self.ori = np.array(ori)
        self.sf = np.array(sf)
        self.phase = np.array(phase)
        self.frame_index = np.array(frame_index)

        # special case
        self._reset_tf()

        # state machine
        self.s_on_v = np.array(s_on_v)
        self.state_time = np.array(s_present_time)
        self.state_cycle = np.array(s_cycle)
        self.state_new_state = np.array(s_new_state)
        self.state_old_state = np.array(s_old_state)
        self.state_elapsed = np.array(s_state_elapsed)
        self.state_trial_type = np.array(s_trial_type)

    def _reset_functions(self):
        time = []
        stim_index = []
        trial_index = []
        photo_state = []
        contrast = []
        pos_x = []
        pos_y = []
        size_x = []
        size_y = []
        frame_index = []

        s_on_v = []
        s_present_time = []
        s_cycle = []
        s_new_state = []
        s_old_state = []
        s_state_elapsed = []
        s_trial_type = []

        with self.stimlog_file.open() as _f:
            for number, line in enumerate(_f):
                line = line.strip()
                if len(line) == 0:
                    continue

                if line.startswith('#'):
                    self._reset_comment_info(line)
                    continue

                part = line.split(',')

                try:
                    code = int(part[0])

                    if code == 10:
                        time.append(float(part[1]))

                        try:
                            stim_value = int(part[2])
                        except ValueError:  # 'None'
                            stim_value = -1
                        stim_index.append(stim_value)

                        trial_index.append(int(part[3]))
                        photo_state.append(int(part[4]))

                        if len(part) > 5:
                            contrast.append(float(part[5]))
                            pos_x.append(int(float(part[6])))
                            pos_y.append(float(part[7]))
                            size_x.append(float(part[8]))
                            size_y.append(int(part[9]))
                            frame_index.append(int(part[10]))
                        else:
                            # empty list item = -1
                            contrast.append(-1)
                            pos_x.append(-1)
                            pos_y.append(-1)
                            size_x.append(-1)
                            size_y.append(-1)
                            frame_index.append(-1)

                    elif code == 20:
                        s_on_v.append(time[-1])  # sync s to v
                        s_present_time.append(int(part[1]) / 1000)
                        s_cycle.append(int(part[2]))
                        s_new_state.append(int(part[3]))
                        s_old_state.append(int(part[4]))
                        s_state_elapsed.append(int(part[5]) / 1000)
                        s_trial_type.append(int(part[6]))

                    else:
                        raise ValueError(f'unknown code : {code}')

                except BaseException as e:
                    raise RuntimeError(f'line {number}: {line}') from e

        self.time = np.array(time)
        self.stim_index = np.array(stim_index)
        self.trial_index = np.array(trial_index)
        self.photo_state = np.array(photo_state)
        self.contrast = np.array(contrast)
        self.pos_x = np.array(pos_x)
        self.pos_y = np.array(pos_y)
        self.size_x = np.array(size_x)
        self.size_y = np.array(size_y)
        self.frame_index = np.array(frame_index)

        # state machine
        self.s_on_v = np.array(s_on_v)
        self.state_time = np.array(s_present_time)
        self.state_cycle = np.array(s_cycle)
        self.state_new_state = np.array(s_new_state)
        self.state_old_state = np.array(s_old_state)
        self.state_elapsed = np.array(s_state_elapsed)
        self.state_trial_type = np.array(s_trial_type)

    def _reset_tf(self):
        try:
            prot = self.riglog_data.get_protocol()
        except AttributeError:
            warnings.warn('cannot inferred prot file due to class not has riglog attribute', vtype='warning')
            pass
        else:
            v_start = self.frame_index == 1
            nr = self.stim_index[v_start]
            self.tf = prot.tf[nr]

    def _reset_comment_info(self, line: str):
        if 'CODES' in line:
            heading, content = line.split(': ')
            iter_codes = content.split(',')
            for pair in iter_codes:
                code, num = pair.split('=')
                code = code.strip()
                value = int(num.strip())
                self.log_info[value] = code
            logger.debug(f'Parsed log_info: {self.log_info}')

        elif 'VLOG HEADER' in line:
            heading, content = line.split(':')
            self.log_header[10] = content.split(', ')
            logger.debug(f'Parsed VLOG HEADER: {self.log_header[10]}')

        elif 'STATE HEADER' in line:
            heading, content = line.split(': ')
            self.log_header[20] = content.split(',')
            logger.debug(f'Parsed STATE HEADER: {self.log_header[20]}')

        elif 'Commit' in line:
            heading, content = line.split(': ')
            commit = content.strip()
            self.config['commit_hash'] = commit
            logger.debug(f'Parsed commit hash: {commit}')

        elif 'Missed' in line:
            match = re.search(r'\d+', line)
            if match:
                missed = int(match.group())
            else:
                missed = None
            self.config['missed_frames'] = missed
            logger.debug(f'Parsed missing frame: {missed}')

    def get_visual_stim_dataframe(self, stim_only: bool = True) -> pl.DataFrame:
        """
        Get the stimlog visual stimulation logging as dataframe::

            ┌─────────────┬───────┬────────┬───────┬───┬───────┬──────┬───────────┬──────────┐
            │ presentTime ┆ iStim ┆ iTrial ┆ photo ┆ … ┆ ori   ┆ sf   ┆ phase     ┆ stim_idx │
            │ ---         ┆ ---   ┆ ---    ┆ ---   ┆   ┆ ---   ┆ ---  ┆ ---       ┆ ---      │
            │ f64         ┆ f64   ┆ f64    ┆ f64   ┆   ┆ f64   ┆ f64  ┆ f64       ┆ f64      │
            ╞═════════════╪═══════╪════════╪═══════╪═══╪═══════╪══════╪═══════════╪══════════╡
            │ 904.048038  ┆ 69.0  ┆ 0.0    ┆ 0.0   ┆ … ┆ 270.0 ┆ 0.16 ┆ 0.066667  ┆ 1.0      │
            │ …           ┆ …     ┆ …      ┆ …     ┆ … ┆ …     ┆ …    ┆ …         ┆ …        │
            │ 2710.807378 ┆ 45.0  ┆ 4.0    ┆ 0.0   ┆ … ┆ 270.0 ┆ 0.04 ┆ 11.933333 ┆ 179.0    │
            │ 2710.824102 ┆ 45.0  ┆ 4.0    ┆ 0.0   ┆ … ┆ 270.0 ┆ 0.04 ┆ 11.933333 ┆ 179.0    │
            └─────────────┴───────┴────────┴───────┴───┴───────┴──────┴───────────┴──────────┘

        :param stim_only: only show the stimulation epoch
        :return: visual stimuli dataframe
        """
        stim_type = self.riglog_data.get_stimulus_type()
        headers = self.log_header[10][1:]
        mask = self.frame_index != -1 if stim_only else slice(None, None)

        match stim_type, self._reset_mapping:
            case ('gratings', None):
                return pl.DataFrame(
                    np.vstack([self.time[mask],
                               self.stim_index[mask],
                               self.trial_index[mask],
                               self.photo_state[mask],
                               self.contrast[mask],
                               self.ori[mask],
                               self.sf[mask],
                               self.phase[mask],
                               self.frame_index[mask]]),
                    schema=headers
                )

            case ('functions', None):
                return pl.DataFrame(
                    np.vstack([self.time[mask],
                               self.stim_index[mask],
                               self.trial_index[mask],
                               self.photo_state[mask],
                               self.contrast[mask],
                               self.pos_x[mask],
                               self.pos_y[mask],
                               self.size_x[mask],
                               self.size_y[mask],
                               self.frame_index[mask]]),
                    schema=headers
                )

            case _:
                return pl.DataFrame(
                    np.vstack([getattr(self, f)[mask] for f in self._reset_mapping[10]]),
                    schema=headers
                )

    def get_state_machine_dataframe(self) -> pl.DataFrame:
        """
        State Machine dataframe::

            ┌──────────┬───────┬──────────┬──────────┬──────────────┬───────────┐
            │ elapsed  ┆ cycle ┆ newState ┆ oldState ┆ stateElapsed ┆ trialType │
            │ ---      ┆ ---   ┆ ---      ┆ ---      ┆ ---          ┆ ---       │
            │ f64      ┆ f64   ┆ f64      ┆ f64      ┆ f64          ┆ f64       │
            ╞══════════╪═══════╪══════════╪══════════╪══════════════╪═══════════╡
            │ 902.601  ┆ 0.0   ┆ 1.0      ┆ 0.0      ┆ 902.601      ┆ 0.0       │
            │ 904.614  ┆ 0.0   ┆ 2.0      ┆ 1.0      ┆ 2.012        ┆ 0.0       │
            │ …        ┆ …     ┆ …        ┆ …        ┆ …            ┆ …         │
            │ 2711.425 ┆ 0.0   ┆ 3.0      ┆ 2.0      ┆ 3.01         ┆ 0.0       │
            │ 2711.425 ┆ 0.0   ┆ 0.0      ┆ 3.0      ┆ 0.0          ┆ 0.0       │
            └──────────┴───────┴──────────┴──────────┴──────────────┴───────────┘

        :return:
        """
        headers = self.log_header[20][1:]

        if self._reset_mapping is None:
            return pl.DataFrame(
                np.vstack([self.state_time,
                           self.state_cycle,
                           self.state_new_state,
                           self.state_old_state,
                           self.state_elapsed,
                           self.state_trial_type]),
                schema=headers
            )
        else:
            return pl.DataFrame(
                np.vstack([getattr(self, f) for f in self._reset_mapping[20]]),
                schema=headers
            )

    @property
    def exp_start_time(self) -> float:
        tstart = self.time[0]

        if isinstance(self.time_offset, float):
            return float(tstart + self.time_offset)
        elif isinstance(self.time_offset, np.ndarray):
            return float(tstart + self.time_offset[0])
        else:
            raise TypeError('')

    @property
    def exp_end_time(self) -> float:
        tend = self.time[-1]

        if isinstance(self.time_offset, float):
            return float(tend + self.time_offset)
        elif isinstance(self.time_offset, np.ndarray):
            return float(tend + self.time_offset[-1])
        else:
            raise TypeError('')

    @property
    def stim_start_time(self) -> float:
        v_start = np.nonzero(self.frame_index == 1)[0][0]
        tstart = self.time[v_start]

        if isinstance(self.time_offset, float):
            return float(tstart + self.time_offset)
        elif isinstance(self.time_offset, np.ndarray):
            return float(tstart + self.time_offset[0])
        else:
            raise TypeError('')

    @property
    def stim_end_time(self) -> float:
        v_end = np.nonzero(np.diff(self.frame_index) < 0)[0][-1] + 1
        tend = self.time[v_end]

        if isinstance(self.time_offset, float):
            return float(tend + self.time_offset)
        elif isinstance(self.time_offset, np.ndarray):
            return float(tend + self.time_offset[-1])
        else:
            raise TypeError('')

    @property
    def stimulus_segment(self) -> np.ndarray:
        v_start = self.frame_index == 1
        t1 = self.time[v_start]
        t2 = self.time[np.nonzero(np.diff(self.frame_index) < 0)[0] + 1]

        if isinstance(self.time_offset, float):
            offset = self.time_offset
        elif isinstance(self.time_offset, np.ndarray):
            offset = np.stack([self.time_offset, self.time_offset], axis=1)
        else:
            raise TypeError('')

        t = np.vstack((t1, t2)).T + offset

        return t

    @property
    def time_offset(self) -> float | np.ndarray:
        if self._cache_time_offset is None:
            self._cache_time_offset = diode_time_offset(
                self.riglog_data,
                self.diode_offset,
                return_sequential=self._do_sequential_diode_offset
            )
        return self._cache_time_offset

    @property
    def time_offset_statemachine(self) -> np.ndarray:
        """time offset approach using statemachine"""
        return self.state_time - self.s_on_v

    def session_trials(self) -> dict[Session, SessionInfo]:
        return {
            prot.name: prot
            for prot in get_protocol_sessions(self)
        }

    def get_stim_pattern(self, with_dur: bool = False) -> AbstractStimulusPattern:
        """
        Get stimulus pattern container

        :param with_dur: if extract theoretical duration value from protocol file
        :return: :class:`~stimpyp.base.AbstractStimulusPattern` based on stim type
        """
        prot = self.riglog_data.get_protocol()
        stim_type = self.riglog_data.get_stimulus_type()
        v_start = self.frame_index == 1
        t = self.stimulus_segment
        contrast = self.contrast[v_start]
        nr = self.stim_index[v_start]

        if stim_type == 'gratings':
            dire = self.ori[v_start]
            sf = self.sf[v_start]
            tf = prot.tf[nr]

            return GratingPattern(t, contrast, dire, sf, tf,
                                  duration=prot['dur'][nr] if with_dur else None)

        elif stim_type == 'functions':
            pos_x = self.pos_x[v_start]
            pos_y = self.pos_y[v_start]
            size_x = self.size_x[v_start]
            size_y = self.size_y[v_start]

            pos_xy = np.vstack([pos_x, pos_y]).T
            size_xy = np.vstack([size_x, size_y]).T

            return FunctionPattern(t, contrast, pos_xy, size_xy,
                                   duration=prot['dur'][nr] if with_dur else None)

        else:
            raise NotImplementedError('')

    @property
    def profile_dataframe(self) -> pl.DataFrame:
        x = np.logical_and(self.stim_index >= 0, self.trial_index != np.max(self.trial_index))  # TOOD check
        stim_index = self.stim_index[x]
        trial_index = self.trial_index[x]
        return pl.DataFrame({
            'i_stims': stim_index.astype(int),
            'i_trials': trial_index.astype(int)
        }).unique(maintain_order=True)


# ================= #
# Diode Time Offset #
# ================= #

class DiodeNumberMismatchError(ValueError):

    def __init__(self):
        super().__init__('Diode numbers are not detected reliably')


class DiodeSignalMissingError(RuntimeError):

    def __init__(self):
        super().__init__('no diode signal were found')


def diode_time_offset(rig: RiglogData,
                      diode_offset: bool = True,
                      return_sequential: bool = True,
                      default_offset_value: float = 0.6) -> float | np.ndarray:
    """
    time offset used in the `old stimpy`
    offset time from screen_time .riglog (diode) to .stimlog
    ** normally stimlog time value are smaller than riglog

    :param rig: :class:`RiglogData`
    :param diode_offset: whether correct diode signal
    :param return_sequential: return sequential offset, if False, use mean value across diode pulses
    :param default_offset_value: hardware(rig)-dependent offset value

    :return: tuple of offset average and std value
    """

    stimlog = rig.get_stimlog()
    if not isinstance(stimlog, Stimlog):
        raise TypeError('')

    if not diode_offset:
        warnings.warn('no offset')
        return default_offset_value

    #
    try:
        t = _diode_offset_sequential(rig, debug_plot=False)
    except DiodeNumberMismatchError as e:
        try:
            first_pulse = _check_if_diode_pulse(rig)
            logger.warning(f'{repr(e)}, use the first pulse diff for alignment')
            return first_pulse

        except DiodeSignalMissingError as e:
            logger.warning(f'{repr(e)}, use default value')
            return default_offset_value

    #
    avg_t = float(np.mean(t))
    std_t = float(np.std(t))

    if not (0 <= avg_t <= 1):
        logger.error(f'{avg_t} too large, might not be properly calculated, check...')

    logger.info(f'DIODE OFFSET avg: {avg_t}s, std: {std_t}s')

    if return_sequential:
        return t
    else:
        return avg_t


def _check_if_diode_pulse(rig: RiglogData) -> float:
    """only count 1st difference"""
    screen_time = rig.screen_event.time
    stimlog = rig.get_stimlog()
    if not isinstance(stimlog, Stimlog):
        raise TypeError('')
    stimlog_vstart = stimlog.time[(stimlog.frame_index == 1)]

    try:
        # noinspection PyTypeChecker
        return screen_time[0] - stimlog_vstart[0]
    except IndexError as e:
        raise DiodeSignalMissingError() from e


def _diode_offset_sequential(rig: RiglogData, debug_plot: bool = False) -> np.ndarray:
    stimlog = rig.get_stimlog()
    if not isinstance(stimlog, Stimlog):
        raise TypeError('')

    stimlog_vstart = stimlog.time[(stimlog.frame_index == 1)]
    riglog_vstart = rig.screen_event.time[0::2]

    if len(riglog_vstart) != len(stimlog_vstart):
        raise DiodeNumberMismatchError()
    else:
        if debug_plot:
            _plot_time_alignment_diode(riglog_vstart, stimlog_vstart)

        return riglog_vstart - stimlog_vstart


def _plot_time_alignment_diode(riglog_screen: np.ndarray,
                               stimlog_time: np.ndarray):
    """Plot time alignment (stimlog time value smaller than riglog)"""
    from matplotlib import pyplot as plt

    fig, ax = plt.subplots()
    ax.plot(riglog_screen - stimlog_time)
    ax.set(xlabel='Visual stim #', ylabel='Time diff (s)')


# ======== #
# Protocol #
# ======== #

def load_protocol(file: PathLike) -> StimpyProtocol:
    """Load stimpy protocol file

    :param file: protocol file path
    """
    return StimpyProtocol.load(file)


class StimpyProtocol(AbstractStimProtocol):
    r"""
    class for handle the protocol file for stimpy **bitbucket/github** version
    (mainly tested in the commits derived from master branch)

    `Dimension parameters`:

        N = numbers of visual stimulation (on-off pairs) = (T \* S)

        T = number of trials

        S = number of Stim Type

        C = number of Cycle
    """

    @classmethod
    def load(cls, file: PathLike) -> Self:
        file = Path(file)
        with file.open() as f:
            return cls._load(file.name, f)

    @classmethod
    def loads(cls, content: str, name: str = '<string>') -> Self:
        """load string"""
        return cls._load(name, content.split('\n'))

    @classmethod
    def _load(cls, name: str, content: Iterable[str]) -> Self:
        options = {}
        version = 'stimpy-bit'

        state = 0
        for line in content:
            line = line.strip()

            if len(line) == 0 or line.startswith('#'):
                continue

            # change state to 1 if # stimulus conditions
            if state == 0 and line.startswith('n '):
                header = re.split(' +', line)  # extract header
                data = [[] for _ in range(len(header))]
                state = 1

            elif state == 0:
                idx = line.index('=')
                value = try_casting_number(line[idx + 1:].strip())
                if isinstance(value, str) and '#' in value:
                    cmt = value.index('#')
                    value = value[:cmt].strip()
                options[line[:idx].strip()] = value

            elif state == 1:
                parts = re.split(' +', line, maxsplit=len(header))
                rows = unfold_stimuli_condition(parts)

                if len(rows) != 1:
                    version = 'stimpy-git'  # determine

                for r in rows:
                    for i, it in enumerate(r):  # for each col
                        data[i].append(it)
            else:
                raise RuntimeError('illegal state')

        assert len(header) == len(data)

        visual_stimuli = {
            field: list(map(str, data[i])) if field == 'evolveParams' else data[i]
            for i, field in enumerate(header)
        }

        # noinspection PyTypeChecker
        return StimpyProtocol(name, options, pl.DataFrame(visual_stimuli), version)

    @property
    def controller(self) -> str:
        """protocol controller name"""
        return self.options['controller']

    @property
    def is_shuffle(self) -> bool:
        return self.options['shuffle'] == 'True'

    @property
    def background(self) -> float:
        return float(self.options.get('background', 0.5))

    @property
    def start_blank_duration(self) -> int:
        return self.options.get('startBlankDuration', 5)

    @property
    def blank_duration(self) -> int:
        return self.options['blankDuration']

    @property
    def trial_blank_duration(self) -> int:
        return 2  # TODO not tested/checked value?

    @property
    def end_blank_duration(self) -> int:
        return self.options.get('endBlankDuration', 5)

    @property
    def trial_duration(self) -> int:
        dur = self['dur']
        return int(np.sum(dur) + len(dur) * self.blank_duration + self.trial_blank_duration)

    @property
    def visual_duration(self) -> int:
        return self.trial_duration * self.n_trials

    @property
    def total_duration(self) -> int:
        return self.start_blank_duration + self.visual_duration + self.end_blank_duration

    @property
    def texture(self) -> str:
        """stimulus texture, {circle, sqr... todo check in stimpy}"""
        return self.options['texture']

    @property
    def mask(self) -> str:
        """todo check in stimpy"""
        return self.options['mask']

    class EvoledParameter(dict[str, Any]):
        """handle the ``evolveParams`` header in the protocol file
        only available in **stimpy-bit** version
        """

        def __init__(self, keys: list[str], data: np.ndarray):
            self.__keys = keys
            self.__data = data  # {'phase':['linear',1]}

        def __len__(self) -> int:
            return len(self.__keys)

        def keys(self):
            return set(self.__keys)

        def __contains__(self, o) -> bool:
            return o in self.__keys

        def __getitem__(self, k: str) -> Any:
            if k not in self.__keys:  # i.e., 'phase'
                raise KeyError()

            # TODO only work for 'linear' case
            ret = []
            for d in self.__data:
                ret.append(d[k][1])

            return np.array(ret)

    def evolve_param_headers(self) -> list[str]:
        """Get parameter header which set in the 'evolveParams, i.g., 'phase'

        :return: header list.

        """
        keys = set()
        data = self['evolveParams']
        for d in data:
            d = eval(d)
            keys.update(d.keys())
        return list(keys)

    @property
    def evolve_params(self) -> EvoledParameter:
        """Get value from 'evolveParams'.

        Examples:

        >>> log: StimpyProtocol

        How many parameters in evolveParams

        >>> len(log.evolve_params)

        List parameter in 'evolveParams'

        >>> log.evolve_params.keys()

        Dose parameter 'phase' in 'evolveParams'?

        >>> 'phase' in log.evolve_params

        Get phase value from 'evolveParams'.

        >>> log.evolve_params['phase']

        :return:
        """
        if self.version == 'stimpy-git':
            raise DeprecationWarning('new stimpy has no evolveParams header')

        data = np.array([eval(it) for it in self['evolveParams']])  # cast back to dict
        return self.EvoledParameter(self.evolve_param_headers(), data)

    @property
    def tf(self) -> np.ndarray:
        if self.version == 'stimpy-bit':
            return self.evolve_params['phase']
        elif self.version == 'stimpy-git':
            return self['tf']
        else:
            raise NotImplementedError('')



import functools
import warnings
from pathlib import Path

import pandas as pd
import polars as pl

__all__ = ['try_casting_number',
           'unfold_stimuli_condition',
           'deprecated_func',
           'printdf',
           'uglob',
           'cls_hasattr']


def try_casting_number(value: str, do_eval: bool = False) -> float | int | str:
    """
    try casting of string to numerical value

    :param value: str
    :param do_eval: do built in eval()
    :return:
    """
    try:
        return int(value)
    except ValueError:
        try:
            return float(value)
        except ValueError:
            if do_eval:
                try:
                    return eval(value)
                except (SyntaxError, NameError):
                    pass

    return value


def unfold_stimuli_condition(parts: list[str]) -> list[list]:
    """
    unfold the numbers of stimuli (each row in prot file) for parsing

    :param parts: list of str foreach row in the visual stimuli dataframe

        e.g., ['1', '3', '0', '0', '1', '0.04', '0', '0', '200', '200', "{'phase':['linear',1]}"]
    """
    ret = []
    if '-' in parts[0]:
        nstim = list(map(int, parts[0].split('-', maxsplit=1)))
        nstim = tuple((nstim[0], nstim[1] + 1))
        ranging = nstim[1] - nstim[0]  # how many n in each row

        for n in range(*nstim):
            parts[0] = f'{n}'
            ext = []
            for i, it in enumerate(parts):
                if '{i}' in it:
                    it = it.replace('{i}', f'{n - 1}%{ranging}')
                if '{t}' in it:
                    raise NotImplementedError('')

                ext.append(try_casting_number(it, do_eval=True))

            ret.append(ext)

    else:
        for i, it in enumerate(parts):
            ret.append(try_casting_number(it, do_eval=True))

        ret = [ret]
    return ret


def deprecated_func(*, new: str | None = None,
                    remarks: str | None = None,
                    removal_version: str = None):
    """Mark deprecated functions.

    :param new: The renamed new usage
    :param remarks: The reason why the function is deprecated
    :param removal_version: Optional version or date when the function is planned to be removed
    """

    def _decorator(f):

        @functools.wraps(f)
        def _deprecated_func(*args, **kwargs):
            msg = _build_deprecated_message(f.__qualname__, new, remarks, removal_version)

            warnings.warn(
                msg,
                DeprecationWarning,
                stacklevel=2
            )

            return f(*args, **kwargs)

        if f.__doc__ is None:
            _deprecated_func.__doc__ = f"DEPRECATED ({remarks}). "
        else:
            _deprecated_func.__doc__ = f"DEPRECATED ({remarks}). " + f.__doc__

        return _deprecated_func

    return _decorator


def _build_deprecated_message(target: str,
                              alternation: str | None = None,
                              remarks: str | None = None,
                              removal: str | None = None) -> str:
    msg = f'{target} is deprecated'

    if removal is not None:
        msg += f' and will be removed in a future release (after version {removal}).'
    else:
        msg += '.'

    if alternation is not None:
        msg += f' Please use "{alternation}" instead.'

    if remarks is not None:
        msg += f' NOTE: {remarks}.'

    return msg


def printdf(df: pl.DataFrame | pd.DataFrame,
            nrows: int | None = None,
            ncols: int | None = None,
            do_print: bool = True,
            **kwargs) -> str:
    """
    print dataframe with given row numbers (polars)
    if isinstance pandas dataframe, print all.

    :param df: polars or pandas dataframe
    :param nrows: number of rows (applicable in polars case)
    :param ncols: number of columns
    :param do_print: do print otherwise, only return the str
    :param kwargs: additional arguments pass the ``pl.Config()``
    :return:
    """

    if isinstance(df, pl.DataFrame):
        with pl.Config(**kwargs) as cfg:
            rows = df.shape[0] if nrows is None else nrows
            cols = df.shape[1] if ncols is None else ncols
            cfg.set_tbl_rows(rows)
            cfg.set_tbl_cols(cols)

            if do_print:
                print(df)

            return df.__repr__()

    elif isinstance(df, pd.DataFrame):
        ret = df.to_markdown()
        print(ret)
        return ret

    else:
        raise TypeError('')


def uglob(directory: Path,
          pattern: str,
          is_dir: bool = False) -> Path:
    """
    Use glob pattern to find the unique file in the directory.

    :param directory: Directory
    :param pattern: Glob pattern
    :param is_dir: Is the pattern point to a directory?
    :return: The unique path
    :raise FileNotFoundError: the unique path is not existed.
    :raise NotADirectoryError: *directory* is not a directory
    :raise RuntimeError: more than one path are found.
    """

    if not directory.exists():
        raise FileNotFoundError(f'{directory} not exit')

    if not directory.is_dir():
        raise NotADirectoryError(f'{directory} is not a directory')

    f = list(directory.glob(pattern))

    if is_dir:
        f = [ff for ff in f if ff.is_dir()]
    else:
        f = [ff for ff in f if not ff.is_dir()]

    if len(f) == 0:
        t = 'directory' if is_dir else 'file'
        raise FileNotFoundError(f'{directory} not have {t} with the pattern: {pattern}')
    elif len(f) == 1:
        return f[0]
    else:
        f.sort()
        t = 'directories' if is_dir else 'files'
        raise RuntimeError(f'multiple {t} were found in {directory} with the pattern {pattern} >>> {f}')


def cls_hasattr(cls: type, attr: str) -> bool:
    """
    Check if attributes in class

    :param cls: The class to check for the attribute.
    :param attr: The name of the attribute to look for within the class and its hierarchy.
    :return: True if the class or any of its parent classes has the specified attribute, False otherwise.
    """
    if attr in getattr(cls, '__annotations__', {}):
        return True

    for c in cls.mro()[1:]:  # Skip the first class as it's already checked
        if attr in getattr(c, '__annotations__', {}):
            return True

    return False



import math
import operator
from random import randint

import numpy as np
from pyparsing import (
    CaselessKeyword,
    Regex,
    Word,
    alphas,
    alphanums,
    Literal,
    Suppress,
    Forward,
    delimitedList,
    Group,
    ParseException
)

__all__ = ['evaluate_string']

EXPR_STACK = []

BNF = None


class EvaluateStringExpressionError(ParseException):
    def __init__(self, iden: str):
        super().__init__(iden, msg=f"invalid identifier '{iden}'")
        self.iden = iden


def bnf_parser():
    """
    expop   :: '^'
    multop  :: '*' | '/'
    addop   :: '+' | '-'
    integer :: ['+' | '-'] '0'..'9'+
    atom    :: PI | E | real | fn '(' expr ')' | '(' expr ')'
    factor  :: atom [ expop factor ]*
    term    :: factor [ multop factor ]*
    expr    :: term [ addop term ]*
    """

    global BNF

    if BNF is None:
        # use CaselessKeyword for e and pi, to avoid accidentally matching
        # functions that start with 'e' or 'pi' (such as 'exp'); Keyword
        # and CaselessKeyword only match whole words
        e = CaselessKeyword("E")
        pi = CaselessKeyword("PI")

        #
        fnumber = Regex(r"[+-]?\d+(?:\.\d*)?(?:[eE][+-]?\d+)?")
        ident = Word(alphas, alphanums + "_$")

        #
        plus, minus, mult, div = map(Literal, "+-*/")
        lpar, rpar = map(Suppress, "()")
        addop = plus | minus
        multop = mult | div
        expop = Literal("^")

        #
        lesser_than = Literal("<")
        greater_than = Literal('>')
        lesser_or_equal = Literal("<=")
        greater_or_equal = Literal(">=")
        compop = lesser_or_equal | greater_or_equal | lesser_than | greater_than

        #
        expr = Forward()
        expr_list = delimitedList(Group(expr))

        # add parse action that replaces the function identifier with a (name, number of args) tuple
        def insert_fn_argcount_tuple(t):
            fn = t.pop(0)
            num_args = len(t[0])
            t.insert(0, (fn, num_args))

        fn_call = (ident + lpar - Group(expr_list) + rpar).setParseAction(
            insert_fn_argcount_tuple
        )
        atom = (
                (addop[...] | compop[...])
                + (
                        (fn_call | pi | e | fnumber | ident).setParseAction(push_first)
                        | Group(lpar + expr + rpar)
                )
        ).setParseAction(push_unary_minus)

        # by defining exponentiation as "atom [ ^ factor ]..." instead of "atom [ ^ atom ]...", we get right-to-left
        # exponents, instead of left-to-right that is, 2^3^2 = 2^(3^2), not (2^3)^2.
        factor = Forward()
        factor <<= atom + (expop + factor).setParseAction(push_first)[...]
        term = factor + (multop + factor).setParseAction(push_first)[...]
        expr <<= term + ((addop | compop) + term).setParseAction(push_first)[...]
        BNF = expr

    return BNF


def push_unary_minus(toks):
    for t in toks:
        if t == "-":
            EXPR_STACK.append("unary -")
        else:
            break


def push_first(toks):
    EXPR_STACK.append(toks[0])


def evaluate_string(s: str, **p) -> str:
    """

    :param s: evaluate string
    :param p:
    :return:
    """
    EXPR_STACK[:] = []

    try:
        bnf_parser().parseString(s, parseAll=True)
    except ParseException as e:
        raise RuntimeError(s) from e

    val = evaluate_stack(EXPR_STACK[:], p)
    return val


epsilon = 1e-12
opn = {
    "+": operator.add,
    "-": operator.sub,
    "*": operator.mul,
    "/": operator.truediv,
    "^": operator.pow,
    "<": operator.lt,
    ">": operator.gt,
    "<=": operator.le,
    ">=": operator.ge
}

fn = {
    "sin": np.sin,
    "cos": np.cos,
    "tan": np.tan,
    "exp": np.exp,
    "abs": np.abs,
    "trunc": lambda it: int(it),
    "round": np.round,
    "randint": randint,
    "sgn": lambda a: -1 if a < -epsilon else 1 if a > epsilon else 0,
    "multiply": lambda a, b: a * b,
    "hypot": math.hypot,
    'all': np.all
}


def evaluate_stack(s, p: dict[str, float]):
    """

    :param s:
    :param p:
    :return:
    """
    op, num_args = s.pop(), 0
    if isinstance(op, tuple):
        op, num_args = op

    if op == "unary -":
        return -evaluate_stack(s, p)
    elif op in opn:
        # note: operands are pushed onto the stack in reverse order
        op2 = evaluate_stack(s, p)
        op1 = evaluate_stack(s, p)
        return opn[op](op1, op2)
    elif op == "PI":
        return math.pi  # 3.1415926535
    elif op == "E":
        return math.e  # 2.718281828
    elif op in fn:
        # note: args are pushed onto the stack in reverse order
        args = reversed([evaluate_stack(s, p) for _ in range(num_args)])
        if op == 'all':
            return fn[op](list(args))
        else:
            return fn[op](*args)
    elif op in p:
        return p[op]
    elif op[0].isalpha():
        raise EvaluateStringExpressionError(op)
    else:
        # try to evaluate as int first, then as float if int fails
        try:
            return int(op)
        except ValueError:
            return float(op)



import logging
from typing import Literal

__all__ = ['set_log_level']

LOG_LEVEL = Literal['DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL']


def set_log_level(level: LOG_LEVEL):
    """Set logging level for stimpyp and its submodules.

    :param level: {'DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL'}
    """
    logger = logging.getLogger("stimpyp")
    numeric_level = getattr(logging, level, logging.INFO)
    logger.setLevel(numeric_level)
    for h in logger.handlers:
        h.setLevel(numeric_level)
